"""
This module defines decorators which Connexion uses to wrap user provided view functions.
"""
from .main import (  # noqa
    ASGIDecorator,
    FlaskDecorator,
    StarletteDecorator,
    WSGIDecorator,
)



import typing as t

from starlette.routing import Router
from starlette.types import ASGIApp, Receive, Scope, Send

Lifespan = t.Callable[[t.Any], t.AsyncContextManager]


class LifespanMiddleware:
    """
    Middleware that adds support for Starlette lifespan handlers
    (https://www.starlette.io/lifespan/).
    """

    def __init__(self, next_app: ASGIApp, *, lifespan: t.Optional[Lifespan]) -> None:
        self.next_app = next_app
        self._lifespan = lifespan
        # Leverage a Starlette Router for lifespan handling only
        self.router = Router(lifespan=lifespan)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        # If no lifespan is registered, pass to next app so it can be handled downstream.
        if scope["type"] == "lifespan" and self._lifespan:
            await self.router(scope, receive, send)
        else:
            await self.next_app(scope, receive, send)



"""
Connexion is a framework that automagically handles HTTP requests based on OpenAPI Specification
(formerly known as Swagger Spec) of your API described in YAML format. Connexion allows you to
write an OpenAPI specification, then maps the endpoints to your Python functions; this makes it
unique, as many tools generate the specification based on your Python code. You can describe your
REST API in as much detail as you want; then Connexion guarantees that it will work as you
specified.
"""

from .apps import AbstractApp  # NOQA
from .apps.asynchronous import AsyncApp
from .datastructures import NoContent  # NOQA
from .exceptions import ProblemException  # NOQA
from .problem import problem  # NOQA
from .resolver import Resolution, Resolver, RestyResolver  # NOQA
from .utils import not_installed_error  # NOQA

try:
    from connexion.apps.flask import FlaskApi, FlaskApp
except ImportError as e:  # pragma: no cover
    _flask_not_installed_error = not_installed_error(
        e, msg="Please install connexion using the 'flask' extra"
    )
    FlaskApi = _flask_not_installed_error  # type: ignore
    FlaskApp = _flask_not_installed_error  # type: ignore

from connexion.apps.asynchronous import AsyncApi, AsyncApp
from connexion.context import request
from connexion.middleware import ConnexionMiddleware

App = FlaskApp
Api = FlaskApi



"""
This module provides general utility functions used within Connexion.
"""

import asyncio
import functools
import importlib
import inspect
import os
import pkgutil
import sys
import typing as t

import yaml
from starlette.routing import compile_path

from connexion.exceptions import TypeValidationError

if t.TYPE_CHECKING:
    from connexion.middleware.main import API


def boolean(s):
    """
    Convert JSON/Swagger boolean value to Python, raise ValueError otherwise

    >>> boolean('true')
    True

    >>> boolean('false')
    False
    """
    if isinstance(s, bool):
        return s
    elif not hasattr(s, "lower"):
        raise ValueError("Invalid boolean value")
    elif s.lower() == "true":
        return True
    elif s.lower() == "false":
        return False
    else:
        raise ValueError("Invalid boolean value")


# https://github.com/swagger-api/swagger-spec/blob/master/versions/2.0.md#data-types
TYPE_MAP: t.Dict[str, t.Any] = {
    "integer": int,
    "number": float,
    "string": str,
    "boolean": boolean,
    "array": list,
    "object": dict,
    "file": lambda x: x,  # Don't cast files
}  # map of swagger types to python types


def make_type(value: t.Any, type_: str, format_: t.Optional[str]) -> t.Any:
    """Cast a value to the type defined in the specification."""
    # In OpenAPI, files are represented with string type and binary format
    if type_ == "string" and format_ == "binary":
        type_ = "file"

    type_func = TYPE_MAP[type_]
    return type_func(value)


def deep_merge(a, b):
    """merges b into a
    in case of conflict the value from b is used
    """
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                deep_merge(a[key], b[key])
            elif a[key] == b[key]:
                pass
            else:
                # b overwrites a
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


def deep_getattr(obj, attr):
    """
    Recurses through an attribute chain to get the ultimate value.
    """

    attrs = attr.split(".")

    return functools.reduce(getattr, attrs, obj)


def deep_get(obj, keys):
    """
    Recurses through a nested object get a leaf value.

    There are cases where the use of inheritance or polymorphism-- the use of allOf or
    oneOf keywords-- will cause the obj to be a list. In this case the keys will
    contain one or more strings containing integers.

    :type obj: list or dict
    :type keys: list of strings
    """
    if not keys:
        return obj

    if isinstance(obj, list):
        return deep_get(obj[int(keys[0])], keys[1:])
    else:
        return deep_get(obj[keys[0]], keys[1:])


def get_function_from_name(function_name):
    """
    Tries to get function by fully qualified name (e.g. "mymodule.myobj.myfunc")

    :type function_name: str
    """
    if function_name is None:
        raise ValueError("Empty function name")

    if "." in function_name:
        module_name, attr_path = function_name.rsplit(".", 1)
    else:
        module_name = ""
        attr_path = function_name

    module = None
    last_import_error = None

    while not module:
        try:
            module = importlib.import_module(module_name)
        except ImportError as import_error:
            last_import_error = import_error
            if "." in module_name:
                module_name, attr_path1 = module_name.rsplit(".", 1)
                attr_path = f"{attr_path1}.{attr_path}"
            else:
                raise
    try:
        function = deep_getattr(module, attr_path)
    except AttributeError:
        if last_import_error:
            raise last_import_error
        else:
            raise
    return function


def is_json_mimetype(mimetype):
    """
    :type mimetype: str
    :rtype: bool
    """
    if mimetype is None:
        return False

    maintype, subtype = mimetype.split("/")  # type: str, str
    if ";" in subtype:
        subtype, parameter = subtype.split(";", maxsplit=1)
    return maintype == "application" and (
        subtype == "json" or subtype.endswith("+json")
    )


def all_json(mimetypes):
    """
    Returns True if all mimetypes are serialized with json

    :type mimetypes: list
    :rtype: bool

    >>> all_json(['application/json'])
    True
    >>> all_json(['application/x.custom+json'])
    True
    >>> all_json([])
    True
    >>> all_json(['application/xml'])
    False
    >>> all_json(['text/json'])
    False
    >>> all_json(['application/json', 'other/type'])
    False
    >>> all_json(['application/json', 'application/x.custom+json'])
    True
    """
    return all(is_json_mimetype(mimetype) for mimetype in mimetypes)


def is_nullable(param_def):
    return param_def.get("schema", param_def).get("nullable", False) or param_def.get(
        "x-nullable", False
    )  # swagger2


def is_null(value):
    if hasattr(value, "strip") and value.strip() in ["null", "None"]:
        return True

    if value is None:
        return True

    return False


def has_coroutine(function, api=None):
    """
    Checks if function is a coroutine.
    If ``function`` is a decorator (has a ``__wrapped__`` attribute)
    this function will also look at the wrapped function.
    """

    def iscorofunc(func):
        iscorofunc = asyncio.iscoroutinefunction(func)
        while not iscorofunc and hasattr(func, "__wrapped__"):
            func = func.__wrapped__
            iscorofunc = asyncio.iscoroutinefunction(func)
        return iscorofunc

    if api is None:
        return iscorofunc(function)

    else:
        return any(iscorofunc(func) for func in (function, api.get_response))


def yamldumper(openapi):
    """
    Returns a nicely-formatted yaml spec.
    :param openapi: a spec dictionary.
    :return: a nicely-formatted, serialized yaml spec.
    """

    def should_use_block(value):
        char_list = (
            "\u000a"  # line feed
            "\u000d"  # carriage return
            "\u001c"  # file separator
            "\u001d"  # group separator
            "\u001e"  # record separator
            "\u0085"  # next line
            "\u2028"  # line separator
            "\u2029"  # paragraph separator
        )
        for c in char_list:
            if c in value:
                return True
        return False

    def my_represent_scalar(self, tag, value, style=None):
        if should_use_block(value):
            style = "|"
        else:
            style = self.default_style

        node = yaml.representer.ScalarNode(tag, value, style=style)
        if self.alias_key is not None:
            self.represented_objects[self.alias_key] = node
        return node

    class NoAnchorDumper(yaml.dumper.SafeDumper):
        """A yaml Dumper that does not replace duplicate entries
        with yaml anchors.
        """

        def ignore_aliases(self, *args):
            return True

    # Dump long lines as "|".
    yaml.representer.SafeRepresenter.represent_scalar = my_represent_scalar

    return yaml.dump(openapi, allow_unicode=True, Dumper=NoAnchorDumper)


def not_installed_error(exc, *, msg=None):  # pragma: no cover
    """Raises the ImportError when the module/object is actually called with a custom message."""

    def _delayed_error(*args, **kwargs):
        if msg is not None:
            raise type(exc)(msg).with_traceback(exc.__traceback__)
        raise exc

    return _delayed_error


def extract_content_type(
    headers: t.Union[t.List[t.Tuple[bytes, bytes]], t.Dict[str, str]]
) -> t.Optional[str]:
    """Extract the mime type and encoding from the content type headers.

    :param headers: Headers from ASGI scope

    :return: The content type if available in headers, otherwise None
    """
    content_type: t.Optional[str] = None

    header_pairs_type = t.Collection[t.Tuple[t.Union[str, bytes], t.Union[str, bytes]]]
    header_pairs: header_pairs_type = headers.items() if isinstance(headers, dict) else headers  # type: ignore
    for key, value in header_pairs:
        # Headers can always be decoded using latin-1:
        # https://stackoverflow.com/a/27357138/4098821
        if isinstance(key, bytes):
            decoded_key: str = key.decode("latin-1")
        else:
            decoded_key = key

        if decoded_key.lower() == "content-type":
            if isinstance(value, bytes):
                content_type = value.decode("latin-1")
            else:
                content_type = value
            break

    return content_type


def split_content_type(
    content_type: t.Optional[str],
) -> t.Tuple[t.Optional[str], t.Optional[str]]:
    """Split the content type in mime_type and encoding. Other parameters are ignored."""
    mime_type, encoding = None, None

    if content_type is None:
        return mime_type, encoding

    # Check for parameters
    if ";" in content_type:
        mime_type, parameters = content_type.split(";", maxsplit=1)

        # Find parameter describing the charset
        prefix = "charset="
        for parameter in parameters.split(";"):
            if parameter.startswith(prefix):
                encoding = parameter[len(prefix) :]
    else:
        mime_type = content_type
    return mime_type, encoding


def coerce_type(param, value, parameter_type, parameter_name=None):
    # TODO: clean up
    TYPE_MAP = {"integer": int, "number": float, "boolean": boolean, "object": dict}

    def make_type(value, type_literal):
        type_func = TYPE_MAP.get(type_literal)
        return type_func(value)

    param_schema = param.get("schema", param)
    if is_nullable(param_schema) and is_null(value):
        return None

    param_type = param_schema.get("type")
    parameter_name = parameter_name if parameter_name else param.get("name")
    if param_type == "array":
        converted_params = []
        if parameter_type == "header":
            value = value.split(",")
        for v in value:
            try:
                converted = make_type(v, param_schema["items"]["type"])
            except (ValueError, TypeError):
                converted = v
            converted_params.append(converted)
        return converted_params
    elif param_type == "object":
        if param_schema.get("properties"):

            def cast_leaves(d, schema):
                if type(d) is not dict:
                    try:
                        return make_type(d, schema["type"])
                    except (ValueError, TypeError):
                        return d
                for k, v in d.items():
                    if k in schema["properties"]:
                        d[k] = cast_leaves(v, schema["properties"][k])
                return d

            return cast_leaves(value, param_schema)
        return value
    else:
        try:
            return make_type(value, param_type)
        except ValueError:
            raise TypeValidationError(param_type, parameter_type, parameter_name)
        except TypeError:
            return value


def get_root_path(import_name: str) -> str:
    """Copied from Flask:
    https://github.com/pallets/flask/blob/836866dc19218832cf02f8b04911060ac92bfc0b/src/flask/helpers.py#L595

    Find the root path of a package, or the path that contains a
    module. If it cannot be found, returns the current working
    directory.
    """
    # Module already imported and has a file attribute. Use that first.
    mod = sys.modules.get(import_name)

    if mod is not None and hasattr(mod, "__file__") and mod.__file__ is not None:
        return os.path.dirname(os.path.abspath(mod.__file__))

    # Next attempt: check the loader.
    loader = pkgutil.get_loader(import_name)

    # Loader does not exist or we're referring to an unloaded main
    # module or a main module without path (interactive sessions), go
    # with the current working directory.
    if loader is None or import_name == "__main__":
        return os.getcwd()

    if hasattr(loader, "get_filename"):
        filepath = loader.get_filename(import_name)  # type: ignore
    else:
        # Fall back to imports.
        __import__(import_name)
        mod = sys.modules[import_name]
        filepath = getattr(mod, "__file__", None)

        # If we don't have a file path it might be because it is a
        # namespace package. In this case pick the root path from the
        # first module that is contained in the package.
        if filepath is None:
            raise RuntimeError(
                "No root path can be found for the provided module"
                f" {import_name!r}. This can happen because the module"
                " came from an import hook that does not provide file"
                " name information or because it's a namespace package."
                " In this case the root path needs to be explicitly"
                " provided."
            )

    # filepath is import_name.py for a module, or __init__.py for a package.
    return os.path.dirname(os.path.abspath(filepath))


def inspect_function_arguments(function: t.Callable) -> t.Tuple[t.List[str], bool]:
    """
    Returns the list of variables names of a function and if it
    accepts keyword arguments.
    """
    parameters = inspect.signature(function).parameters
    bound_arguments = [
        name
        for name, p in parameters.items()
        if p.kind not in (p.VAR_POSITIONAL, p.VAR_KEYWORD)
    ]
    has_kwargs = any(p.kind == p.VAR_KEYWORD for p in parameters.values())
    return list(bound_arguments), has_kwargs


T = t.TypeVar("T")


@t.overload
def sort_routes(routes: t.List[str], *, key: None = None) -> t.List[str]:
    ...


@t.overload
def sort_routes(routes: t.List[T], *, key: t.Callable[[T], str]) -> t.List[T]:
    ...


def sort_routes(routes, *, key=None):
    """Sorts a list of routes from most specific to least specific.

    See Starlette routing documentation and implementation as this function
    is aimed to sort according to that logic.
    - https://www.starlette.io/routing/#route-priority

    The only difference is that a `path` component is appended to each route
    such that `/` is less specific than `/basepath` while they are technically
    not comparable.
    This is because it is also done by the `Mount` class internally:
    https://github.com/encode/starlette/blob/1c1043ca0ab7126419948b27f9d0a78270fd74e6/starlette/routing.py#L388

    For example, from most to least specific:
    - /users/me
    - /users/{username}/projects/{project}
    - /users/{username}

    :param routes: List of routes to sort
    :param key: Function to extract the path from a route if it is not a string

    :return: List of routes sorted from most specific to least specific
    """

    class SortableRoute:
        def __init__(self, path: str) -> None:
            self.path = path.rstrip("/")
            if not self.path.endswith("/{path:path}"):
                self.path += "/{path:path}"
            self.path_regex, _, _ = compile_path(self.path)

        def __lt__(self, other: "SortableRoute") -> bool:
            return bool(other.path_regex.match(self.path))

    return sorted(routes, key=lambda r: SortableRoute(key(r) if key else r))


def sort_apis_by_basepath(apis: t.List["API"]) -> t.List["API"]:
    """Sorts a list of APIs by basepath.

    :param apis: List of APIs to sort

    :return: List of APIs sorted by basepath
    """
    return sort_routes(apis, key=lambda api: api.base_path or "/")



"""
This module contains resolvers, functions that resolves the user defined view functions
from the operations defined in the OpenAPI spec.
"""

import inspect
import logging
import typing as t

from inflection import camelize

import connexion.utils as utils
from connexion.exceptions import ResolverError

logger = logging.getLogger("connexion.resolver")


class Resolution:
    def __init__(self, function, operation_id):
        """
        Represents the result of operation resolution

        :param function: The endpoint function
        :type function: types.FunctionType
        """
        self.function = function
        self.operation_id = operation_id


class Resolver:
    def __init__(self, function_resolver: t.Callable = utils.get_function_from_name):
        """
        Standard resolver

        :param function_resolver: Function that resolves functions using an operationId
        """
        self.function_resolver = function_resolver

    def resolve(self, operation):
        """
        Default operation resolver

        :type operation: connexion.operations.AbstractOperation
        """
        operation_id = self.resolve_operation_id(operation)
        return Resolution(
            self.resolve_function_from_operation_id(operation_id), operation_id
        )

    def resolve_operation_id(self, operation):
        """
        Default operationId resolver

        :type operation: connexion.operations.AbstractOperation
        """
        operation_id = operation.operation_id
        router_controller = operation.router_controller
        if router_controller is None:
            return operation_id
        return f"{router_controller}.{operation_id}"

    def resolve_function_from_operation_id(self, operation_id):
        """
        Invokes the function_resolver

        :type operation_id: str
        """
        try:
            return self.function_resolver(operation_id)
        except ImportError as e:
            msg = f'Cannot resolve operationId "{operation_id}"! Import error was "{str(e)}"'
            raise ResolverError(msg)
        except (AttributeError, ValueError) as e:
            raise ResolverError(str(e))


class RelativeResolver(Resolver):
    """
    Resolves endpoint functions relative to a given root path or module.
    """

    def __init__(self, root_path, function_resolver=utils.get_function_from_name):
        """
        :param root_path: The root path relative to which an operationId is resolved.
            Can also be a module. Has the same effect as setting
            `x-swagger-router-controller` or `x-openapi-router-controller` equal to
            `root_path` for every operation individually.
        :type root_path: typing.Union[str, types.ModuleType]
        :param function_resolver: Function that resolves functions using an operationId
        :type function_resolver: types.FunctionType
        """
        super().__init__(function_resolver=function_resolver)
        if inspect.ismodule(root_path):
            self.root_path = root_path.__name__
        else:
            self.root_path = root_path

    def resolve_operation_id(self, operation):
        """Resolves the operationId relative to the root path, unless
        x-swagger-router-controller or x-openapi-router-controller is specified.

        :param operation: The operation to resolve
        :type operation: connexion.operations.AbstractOperation
        """
        operation_id = operation.operation_id
        router_controller = operation.router_controller
        if router_controller is None:
            return f"{self.root_path}.{operation_id}"
        return f"{router_controller}.{operation_id}"


class RestyResolver(Resolver):
    """
    Resolves endpoint functions using REST semantics (unless overridden by specifying operationId)
    """

    def __init__(
        self, default_module_name: str, *, collection_endpoint_name: str = "search"
    ):
        """
        :param default_module_name: Default module name for operations
        :param collection_endpoint_name: Name of function to resolve collection endpoints to
        """
        super().__init__()
        self.default_module_name = default_module_name
        self.collection_endpoint_name = collection_endpoint_name

    def resolve_operation_id(self, operation):
        """
        Resolves the operationId using REST semantics unless explicitly configured in the spec

        :type operation: connexion.operations.AbstractOperation
        """
        if operation.operation_id:
            return super().resolve_operation_id(operation)

        return self.resolve_operation_id_using_rest_semantics(operation)

    def resolve_operation_id_using_rest_semantics(self, operation):
        """
        Resolves the operationId using REST semantics

        :type operation: connexion.operations.AbstractOperation
        """

        # Split the path into components delimited by '/'
        path_components = [c for c in operation.path.split("/") if len(c)]

        def is_var(component):
            """True if the path component is a var. eg, '{id}'"""
            return (component[0] == "{") and (component[-1] == "}")

        resource_name = ".".join([c for c in path_components if not is_var(c)]).replace(
            "-", "_"
        )

        def get_controller_name():
            x_router_controller = operation.router_controller

            name = self.default_module_name

            if x_router_controller:
                name = x_router_controller

            elif resource_name:
                name += "." + resource_name

            return name

        def get_function_name():
            method = operation.method

            is_collection_endpoint = (
                method.lower() == "get"
                and len(resource_name)
                and not is_var(path_components[-1])
            )

            return (
                self.collection_endpoint_name
                if is_collection_endpoint
                else method.lower()
            )

        return f"{get_controller_name()}.{get_function_name()}"


class MethodResolverBase(RestyResolver):
    """
    Resolves endpoint functions based on Flask's MethodView semantics, e.g.

    .. code-block:: yaml

        paths:
            /foo_bar:
                get:
                    # Implied function call: api.FooBarView().get

    .. code-block:: python

        class FooBarView(MethodView):
            def get(self):
                return ...
            def post(self):
                return ...

    """

    _class_arguments_type = t.Dict[
        str, t.Dict[str, t.Union[t.Iterable, t.Dict[str, t.Any]]]
    ]

    def __init__(self, *args, class_arguments: _class_arguments_type = None, **kwargs):
        """
        :param args: Arguments passed to :class:`~RestyResolver`
        :param class_arguments: Arguments to instantiate the View Class in the format below
        :param kwargs: Keywords arguments passed to :class:`~RestyResolver`

        .. code-block:: python

            {
              "ViewName": {
                "args": (positional arguments,)
                "kwargs": {
                  "keyword": "argument"
                }
              }
            }
        """
        self.class_arguments = class_arguments or {}
        super(MethodResolverBase, self).__init__(*args, **kwargs)
        self.initialized_views: list = []

    def resolve_operation_id(self, operation):
        """
        Resolves the operationId using REST semantics unless explicitly configured in the spec
        Once resolved with REST semantics the view_name is capitalised and has 'View' added
        to it so it now matches the Class names of the MethodView

        :type operation: connexion.operations.AbstractOperation
        """
        if operation.operation_id:
            # If operation_id is defined then use the higher level API to resolve
            return RestyResolver.resolve_operation_id(self, operation)

        # Use RestyResolver to get operation_id for us (follow their naming conventions/structure)
        operation_id = self.resolve_operation_id_using_rest_semantics(operation)
        module_name, view_base, meth_name = operation_id.rsplit(".", 2)
        view_name = camelize(view_base) + "View"

        return f"{module_name}.{view_name}.{meth_name}"

    def resolve_function_from_operation_id(self, operation_id):
        """
        Invokes the function_resolver

        :type operation_id: str
        """

        try:
            module_name, view_name, meth_name = operation_id.rsplit(".", 2)
            if operation_id and not view_name.endswith("View"):
                # If operation_id is not a view then assume it is a standard function
                return self.function_resolver(operation_id)

            mod = __import__(module_name, fromlist=[view_name])
            view_cls = getattr(mod, view_name)
            # find the view and return it
            return self.resolve_method_from_class(view_name, meth_name, view_cls)

        except ImportError as e:
            msg = 'Cannot resolve operationId "{}"! Import error was "{}"'.format(
                operation_id, str(e)
            )
            raise ResolverError(msg)
        except (AttributeError, ValueError) as e:
            raise ResolverError(str(e))

    def resolve_method_from_class(self, view_name, meth_name, view_cls):
        """
        Returns the view function for the given view class.
        """
        raise NotImplementedError()


class MethodResolver(MethodResolverBase):
    """
    A generic method resolver that instantiates a class and extracts the method
    from it, based on the operation id.
    """

    def resolve_method_from_class(self, view_name, meth_name, view_cls):
        view = None
        for v in self.initialized_views:
            if v.__class__ == view_cls:
                view = v
                break
        if view is None:
            # get the args and kwargs for this view
            cls_arguments = self.class_arguments.get(view_name, {})
            cls_args = cls_arguments.get("args", ())
            cls_kwargs = cls_arguments.get("kwargs", {})
            # instantiate the class with the args and kwargs
            view = view_cls(*cls_args, **cls_kwargs)
            self.initialized_views.append(view)
        # get the method if the class
        func = getattr(view, meth_name)
        # Return the method function of the class
        return func


class MethodViewResolver(MethodResolverBase):
    """
    A specialized method resolver that works with flask's method views.
    It resolves the method by calling as_view on the class.
    """

    def __init__(self, *args, **kwargs):
        if "collection_endpoint_name" in kwargs:
            del kwargs["collection_endpoint_name"]
            # Dispatch of request is done by Flask
            logger.warning(
                "collection_endpoint_name is ignored by the MethodViewResolver. "
                "Requests to a collection endpoint will be routed to .get()"
            )
        super().__init__(*args, **kwargs)

    def resolve_method_from_class(self, view_name, meth_name, view_cls):
        view = None
        for v in self.initialized_views:
            # views returned by <class>.as_view
            # have the origin class attached as .view_class
            if v.view_class == view_cls:
                view = v
                break
        if view is None:
            # get the args and kwargs for this view
            cls_arguments = self.class_arguments.get(view_name, {})
            cls_args = cls_arguments.get("args", ())
            cls_kwargs = cls_arguments.get("kwargs", {})
            # call as_view to get a view function
            # that is decorated with the classes
            # decorator list, if any
            view = view_cls.as_view(view_name, *cls_args, **cls_kwargs)
            # add the view to the list of initialized views
            # in order to call as_view only once
            self.initialized_views.append(view)
        # return the class as view function
        # for each operation so that requests
        # are dispatched with <class>.dispatch_request,
        # when calling the view function
        return view



"""
Validation Middleware.
"""
import logging
import typing as t

from starlette.types import ASGIApp, Receive, Scope, Send

from connexion import utils
from connexion.datastructures import MediaTypeDict
from connexion.exceptions import NonConformingResponseHeaders
from connexion.middleware.abstract import RoutedAPI, RoutedMiddleware
from connexion.operations import AbstractOperation
from connexion.validators import VALIDATOR_MAP

logger = logging.getLogger("connexion.middleware.validation")


class ResponseValidationOperation:
    def __init__(
        self,
        next_app: ASGIApp,
        *,
        operation: AbstractOperation,
        validator_map: t.Optional[dict] = None,
    ) -> None:
        self.next_app = next_app
        self._operation = operation
        self._validator_map = VALIDATOR_MAP.copy()
        self._validator_map.update(validator_map or {})

    def extract_content_type(
        self, headers: t.List[t.Tuple[bytes, bytes]]
    ) -> t.Tuple[str, str]:
        """Extract the mime type and encoding from the content type headers.

        :param headers: Headers from ASGI scope

        :return: A tuple of mime type, encoding
        """
        content_type = utils.extract_content_type(headers)
        mime_type, encoding = utils.split_content_type(content_type)
        if mime_type is None:
            # Content-type header is not required. Take a best guess.
            try:
                mime_type = self._operation.produces[0]
            except IndexError:
                mime_type = "application/octet-stream"
        if encoding is None:
            encoding = "utf-8"

        return mime_type, encoding

    def validate_mime_type(self, mime_type: str) -> None:
        """Validate the mime type against the spec if it defines which mime types are produced.

        :param mime_type: mime type from content type header
        """
        if not self._operation.produces:
            return

        media_type_dict = MediaTypeDict(
            [(p.lower(), None) for p in self._operation.produces]
        )
        if mime_type.lower() not in media_type_dict:
            raise NonConformingResponseHeaders(
                detail=f"Invalid Response Content-type ({mime_type}), "
                f"expected {self._operation.produces}",
            )

    @staticmethod
    def validate_required_headers(
        headers: t.List[tuple], response_definition: dict
    ) -> None:
        required_header_keys = {
            k.lower()
            for (k, v) in response_definition.get("headers", {}).items()
            if v.get("required", False)
        }
        header_keys = set(header[0].decode("latin-1").lower() for header in headers)
        missing_keys = required_header_keys - header_keys
        if missing_keys:
            pretty_list = ", ".join(missing_keys)
            msg = (
                "Keys in response header don't match response specification. Difference: {}"
            ).format(pretty_list)
            raise NonConformingResponseHeaders(detail=msg)

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        async def wrapped_send(message: t.MutableMapping[str, t.Any]) -> None:
            nonlocal send

            if message["type"] == "http.response.start":
                headers = message["headers"]

                mime_type, encoding = self.extract_content_type(headers)
                if message["status"] < 400:
                    self.validate_mime_type(mime_type)

                status = str(message["status"])
                response_definition = self._operation.response_definition(
                    status, mime_type
                )
                self.validate_required_headers(headers, response_definition)

                # Validate body
                try:
                    body_validator = self._validator_map["response"][mime_type]  # type: ignore
                except KeyError:
                    logging.info(
                        f"Skipping validation. No validator registered for content type: "
                        f"{mime_type}."
                    )
                else:
                    validator = body_validator(
                        scope,
                        schema=self._operation.response_schema(status, mime_type),
                        nullable=utils.is_nullable(
                            self._operation.response_definition(status, mime_type)
                        ),
                        encoding=encoding,
                    )
                    send = validator.wrap_send(send)

            return await send(message)

        await self.next_app(scope, receive, wrapped_send)


class ResponseValidationAPI(RoutedAPI[ResponseValidationOperation]):
    """Validation API."""

    def __init__(
        self,
        *args,
        validator_map=None,
        validate_responses=False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.validator_map = validator_map
        self.validate_responses = validate_responses
        self.add_paths()

    def make_operation(
        self, operation: AbstractOperation
    ) -> ResponseValidationOperation:
        if self.validate_responses:
            return ResponseValidationOperation(
                self.next_app,
                operation=operation,
                validator_map=self.validator_map,
            )
        else:
            return self.next_app  # type: ignore


class ResponseValidationMiddleware(RoutedMiddleware[ResponseValidationAPI]):
    """Middleware for validating requests according to the API contract."""

    api_cls = ResponseValidationAPI



import abc
import logging
import typing as t
from collections import defaultdict

from starlette.types import ASGIApp, Receive, Scope, Send

from connexion.exceptions import MissingMiddleware, ResolverError
from connexion.http_facts import METHODS
from connexion.operations import AbstractOperation
from connexion.resolver import Resolver
from connexion.spec import Specification

logger = logging.getLogger(__name__)

ROUTING_CONTEXT = "connexion_routing"


class SpecMiddleware(abc.ABC):
    """Middlewares that need the specification(s) to be registered on them should inherit from this
    base class"""

    @abc.abstractmethod
    def add_api(self, specification: Specification, **kwargs) -> t.Any:
        """
        Register an API represented by a single OpenAPI specification on this middleware.
        Multiple APIs can be registered on a single middleware.
        """

    @abc.abstractmethod
    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        pass


class AbstractSpecAPI:
    """Base API class with only minimal behavior related to the specification."""

    def __init__(
        self,
        specification: Specification,
        base_path: t.Optional[str] = None,
        resolver: t.Optional[Resolver] = None,
        uri_parser_class=None,
        *args,
        **kwargs,
    ):
        self.specification = specification
        self.uri_parser_class = uri_parser_class

        self._set_base_path(base_path)

        self.resolver = resolver or Resolver()

    def _set_base_path(self, base_path: t.Optional[str] = None) -> None:
        if base_path is not None:
            # update spec to include user-provided base_path
            self.specification.base_path = base_path
            self.base_path = base_path
        else:
            self.base_path = self.specification.base_path


OP = t.TypeVar("OP")
"""Typevar representing an operation"""


class AbstractRoutingAPI(AbstractSpecAPI, t.Generic[OP]):
    """Base API class with shared functionality related to routing."""

    def __init__(
        self,
        *args,
        pythonic_params=False,
        resolver_error_handler: t.Optional[t.Callable] = None,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.pythonic_params = pythonic_params
        self.resolver_error_handler = resolver_error_handler

        self.add_paths()

    def add_paths(self, paths: t.Optional[dict] = None) -> None:
        """
        Adds the paths defined in the specification as operations.
        """
        paths = t.cast(dict, paths or self.specification.get("paths", dict()))
        for path, methods in paths.items():
            logger.debug("Adding %s%s...", self.base_path, path)

            for method in methods:
                if method not in METHODS:
                    continue
                try:
                    self.add_operation(path, method)
                except ResolverError as err:
                    # If we have an error handler for resolver errors, add it as an operation.
                    # Otherwise treat it as any other error.
                    if self.resolver_error_handler is not None:
                        self._add_resolver_error_handler(method, path, err)
                    else:
                        self._handle_add_operation_error(path, method, err)
                except Exception as e:
                    # All other relevant exceptions should be handled as well.
                    self._handle_add_operation_error(path, method, e)

    def add_operation(self, path: str, method: str) -> None:
        """
        Adds one operation to the api.

        This method uses the OperationID identify the module and function that will handle the operation

        From Swagger Specification:

        **OperationID**

        A friendly name for the operation. The id MUST be unique among all operations described in the API.
        Tools and libraries MAY use the operation id to uniquely identify an operation.
        """
        spec_operation_cls = self.specification.operation_cls
        spec_operation = spec_operation_cls.from_spec(
            self.specification,
            path=path,
            method=method,
            resolver=self.resolver,
            uri_parser_class=self.uri_parser_class,
        )
        operation = self.make_operation(spec_operation)
        path, name = self._framework_path_and_name(spec_operation, path)
        self._add_operation_internal(method, path, operation, name=name)

    @abc.abstractmethod
    def make_operation(self, operation: AbstractOperation) -> OP:
        """Build an operation to register on the API."""

    @staticmethod
    def _framework_path_and_name(
        operation: AbstractOperation, path: str
    ) -> t.Tuple[str, str]:
        """Prepare the framework path & name to register the operation on the API."""

    @abc.abstractmethod
    def _add_operation_internal(
        self, method: str, path: str, operation: OP, name: str = None
    ) -> None:
        """
        Adds the operation according to the user framework in use.
        It will be used to register the operation on the user framework router.
        """

    def _add_resolver_error_handler(
        self, method: str, path: str, err: ResolverError
    ) -> None:
        """
        Adds a handler for ResolverError for the given method and path.
        """
        self.resolver_error_handler = t.cast(t.Callable, self.resolver_error_handler)
        operation = self.resolver_error_handler(
            err,
        )
        self._add_operation_internal(method, path, operation)

    def _handle_add_operation_error(
        self, path: str, method: str, exc: Exception
    ) -> None:
        url = f"{self.base_path}{path}"
        error_msg = f"Failed to add operation for {method.upper()} {url}"
        logger.error(error_msg)
        raise exc from None


class RoutedAPI(AbstractSpecAPI, t.Generic[OP]):
    def __init__(
        self,
        specification: Specification,
        *args,
        next_app: ASGIApp,
        **kwargs,
    ) -> None:
        super().__init__(specification, *args, **kwargs)
        self.next_app = next_app
        self.operations: t.MutableMapping[t.Optional[str], OP] = {}

    def add_paths(self) -> None:
        paths = self.specification.get("paths", {})
        for path, methods in paths.items():
            for method in methods:
                if method not in METHODS:
                    continue
                try:
                    self.add_operation(path, method)
                except ResolverError:
                    # ResolverErrors are either raised or handled in routing middleware.
                    pass

    def add_operation(self, path: str, method: str) -> None:
        operation_spec_cls = self.specification.operation_cls
        operation = operation_spec_cls.from_spec(
            self.specification,
            path=path,
            method=method,
            resolver=self.resolver,
            uri_parser_class=self.uri_parser_class,
        )
        routed_operation = self.make_operation(operation)
        self.operations[operation.operation_id] = routed_operation

    @abc.abstractmethod
    def make_operation(self, operation: AbstractOperation) -> OP:
        """Create an operation of the `operation_cls` type."""
        raise NotImplementedError


API = t.TypeVar("API", bound="RoutedAPI")
"""Typevar representing an API which subclasses RoutedAPI"""


class RoutedMiddleware(SpecMiddleware, t.Generic[API]):
    """Baseclass for middleware that wants to leverage the RoutingMiddleware to route requests to
    its operations.

    The RoutingMiddleware adds the operation_id to the ASGI scope. This middleware registers its
    operations by operation_id at startup. At request time, the operation is fetched by an
    operation_id lookup.
    """

    api_cls: t.Type[API]
    """The subclass of RoutedAPI this middleware uses."""

    def __init__(self, app: ASGIApp) -> None:
        self.app = app
        self.apis: t.Dict[str, t.List[API]] = defaultdict(list)

    def add_api(self, specification: Specification, **kwargs) -> API:
        api = self.api_cls(specification, next_app=self.app, **kwargs)
        self.apis[api.base_path].append(api)
        return api

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """Fetches the operation related to the request and calls it."""
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        try:
            connexion_context = scope["extensions"][ROUTING_CONTEXT]
        except KeyError:
            raise MissingMiddleware(
                "Could not find routing information in scope. Please make sure "
                "you have a routing middleware registered upstream. "
            )
        api_base_path = connexion_context.get("api_base_path")
        if api_base_path is not None and api_base_path in self.apis:
            for api in self.apis[api_base_path]:
                operation_id = connexion_context.get("operation_id")
                try:
                    operation = api.operations[operation_id]
                except KeyError:
                    if operation_id is None:
                        logger.debug("Skipping operation without id.")
                        await self.app(scope, receive, send)
                        return
                else:
                    return await operation(scope, receive, send)

            raise MissingOperation("Encountered unknown operation_id.")

        await self.app(scope, receive, send)


class MissingOperation(Exception):
    """Missing operation"""



import json
import logging
import typing as t

import jsonschema
from jsonschema import Draft4Validator, ValidationError, draft4_format_checker
from starlette.types import Scope

from connexion.exceptions import BadRequestProblem, NonConformingResponseBody
from connexion.json_schema import (
    Draft4RequestValidator,
    Draft4ResponseValidator,
    format_error_with_path,
)
from connexion.validators import (
    AbstractRequestBodyValidator,
    AbstractResponseBodyValidator,
)

logger = logging.getLogger(__name__)


class JSONRequestBodyValidator(AbstractRequestBodyValidator):
    """Request body validator for json content types."""

    def __init__(
        self,
        *,
        schema: dict,
        required=False,
        nullable=False,
        encoding: str,
        strict_validation: bool,
        **kwargs,
    ) -> None:
        super().__init__(
            schema=schema,
            required=required,
            nullable=nullable,
            encoding=encoding,
            strict_validation=strict_validation,
        )

    @property
    def _validator(self):
        return Draft4RequestValidator(
            self._schema, format_checker=draft4_format_checker
        )

    async def _parse(
        self, stream: t.AsyncGenerator[bytes, None], scope: Scope
    ) -> t.Any:
        bytes_body = b"".join([message async for message in stream])
        body = bytes_body.decode(self._encoding)

        if not body:
            return None

        try:
            return json.loads(body)
        except json.decoder.JSONDecodeError as e:
            raise BadRequestProblem(detail=str(e))

    def _validate(self, body: t.Any) -> t.Optional[dict]:
        if not self._nullable and body is None:
            raise BadRequestProblem("Request body must not be empty")
        try:
            return self._validator.validate(body)
        except ValidationError as exception:
            error_path_msg = format_error_with_path(exception=exception)
            logger.error(
                f"Validation error: {exception.message}{error_path_msg}",
                extra={"validator": "body"},
            )
            raise BadRequestProblem(detail=f"{exception.message}{error_path_msg}")


class DefaultsJSONRequestBodyValidator(JSONRequestBodyValidator):
    """Request body validator for json content types which fills in default values. This Validator
    intercepts the body, makes changes to it, and replays it for the next ASGI application."""

    MUTABLE_VALIDATION = True
    """This validator might mutate to the body."""

    @property
    def _validator(self):
        validator_cls = self.extend_with_set_default(Draft4RequestValidator)
        return validator_cls(self._schema, format_checker=draft4_format_checker)

    # via https://python-jsonschema.readthedocs.io/
    @staticmethod
    def extend_with_set_default(validator_class):
        validate_properties = validator_class.VALIDATORS["properties"]

        def set_defaults(validator, properties, instance, schema):
            for property, subschema in properties.items():
                if "default" in subschema:
                    instance.setdefault(property, subschema["default"])

            yield from validate_properties(validator, properties, instance, schema)

        return jsonschema.validators.extend(
            validator_class, {"properties": set_defaults}
        )


class JSONResponseBodyValidator(AbstractResponseBodyValidator):
    """Response body validator for json content types."""

    @property
    def validator(self) -> Draft4Validator:
        return Draft4ResponseValidator(
            self._schema, format_checker=draft4_format_checker
        )

    def _parse(self, stream: t.Generator[bytes, None, None]) -> t.Any:
        body = b"".join(stream).decode(self._encoding)

        if not body:
            return None

        try:
            return json.loads(body)
        except json.decoder.JSONDecodeError as e:
            raise NonConformingResponseBody(str(e))

    def _validate(self, body: dict):
        try:
            self.validator.validate(body)
        except ValidationError as exception:
            error_path_msg = format_error_with_path(exception=exception)
            logger.error(
                f"Validation error: {exception.message}{error_path_msg}",
                extra={"validator": "body"},
            )
            raise NonConformingResponseBody(
                detail=f"Response body does not conform to specification. {exception.message}{error_path_msg}"
            )


class TextResponseBodyValidator(JSONResponseBodyValidator):
    def _parse(self, stream: t.Generator[bytes, None, None]) -> str:  # type: ignore
        body = b"".join(stream).decode(self._encoding)

        try:
            return json.loads(body)
        except json.decoder.JSONDecodeError:
            return body



"""
This module defines a Swagger2Operation class, a Connexion operation specific for Swagger 2 specs.
"""

import logging
import typing as t

from connexion.exceptions import InvalidSpecification
from connexion.operations.abstract import AbstractOperation
from connexion.uri_parsing import Swagger2URIParser
from connexion.utils import deep_get

logger = logging.getLogger("connexion.operations.swagger2")


COLLECTION_FORMAT_MAPPING = {
    "multi": {"style": "form", "explode": True},
    "csv": {"style": "form", "explode": False},
    "ssv": {"style": "spaceDelimited", "explode": False},
    "pipes": {"style": "pipeDelimited", "explode": False},
}


class Swagger2Operation(AbstractOperation):

    """
    Exposes a Swagger 2.0 operation under the AbstractOperation interface.
    The primary purpose of this class is to provide the `function()` method
    to the API. A Swagger2Operation is plugged into the API with the provided
    (path, method) pair. It resolves the handler function for this operation
    with the provided resolver, and wraps the handler function with multiple
    decorators that provide security, validation, serialization,
    and deserialization.
    """

    def __init__(
        self,
        method,
        path,
        operation,
        resolver,
        app_produces,
        app_consumes,
        path_parameters=None,
        app_security=None,
        security_schemes=None,
        definitions=None,
        randomize_endpoint=None,
        uri_parser_class=None,
    ):
        """
        :param method: HTTP method
        :type method: str
        :param path: relative path to this operation
        :type path: str
        :param operation: swagger operation object
        :type operation: dict
        :param resolver: Callable that maps operationID to a function
        :type resolver: resolver.Resolver
        :param app_produces: list of content types the application can return by default
        :type app_produces: list
        :param app_consumes: list of content types the application consumes by default
        :type app_consumes: list
        :param path_parameters: Parameters defined in the path level
        :type path_parameters: list
        :param app_security: list of security rules the application uses by default
        :type app_security: list
        :param security_schemes: `Security Definitions Object
            <https://github.com/swagger-api/swagger-spec/blob/master/versions/2.0.md#security-definitions-object>`_
        :type security_schemes: dict
        :param definitions: `Definitions Object
            <https://github.com/swagger-api/swagger-spec/blob/master/versions/2.0.md#definitionsObject>`_
        :type definitions: dict
        :param randomize_endpoint: number of random characters to append to operation name
        :type randomize_endpoint: integer
        :param uri_parser_class: class to use for uri parsing
        :type uri_parser_class: AbstractURIParser
        """
        uri_parser_class = uri_parser_class or Swagger2URIParser

        self._router_controller = operation.get("x-swagger-router-controller")

        super().__init__(
            method=method,
            path=path,
            operation=operation,
            resolver=resolver,
            app_security=app_security,
            security_schemes=security_schemes,
            randomize_endpoint=randomize_endpoint,
            uri_parser_class=uri_parser_class,
        )

        self._produces = operation.get("produces", app_produces)
        self._consumes = operation.get("consumes", app_consumes)

        self.definitions = definitions or {}

        self._parameters = operation.get("parameters", [])
        if path_parameters:
            self._parameters += path_parameters

        self._responses = operation.get("responses", {})

    @classmethod
    def from_spec(cls, spec, *args, path, method, resolver, **kwargs):
        return cls(
            method,
            path,
            spec.get_operation(path, method),
            resolver=resolver,
            path_parameters=spec.get_path_params(path),
            app_produces=spec.produces,
            app_consumes=spec.consumes,
            app_security=spec.security,
            security_schemes=spec.security_schemes,
            definitions=spec.definitions,
            *args,
            **kwargs,
        )

    @property
    def request_body(self) -> dict:
        if not hasattr(self, "_request_body"):
            body_params = []
            form_params = []
            for parameter in self.parameters:
                if parameter["in"] == "body":
                    body_params.append(parameter)
                elif parameter["in"] == "formData":
                    form_params.append(parameter)

            if len(body_params) > 1:
                raise InvalidSpecification(
                    f"{self.method} {self.path}: There can be one 'body' parameter at most"
                )

            if body_params and form_params:
                raise InvalidSpecification(
                    f"{self.method} {self.path}: 'body' and 'formData' parameters are mutually exclusive"
                )

            if body_params:
                self._request_body = self._transform_json(body_params[0])
            elif form_params:
                self._request_body = self._transform_form(form_params)
            else:
                self._request_body = {}

        return self._request_body

    @property
    def parameters(self):
        return self._parameters

    @property
    def consumes(self):
        return self._consumes

    @property
    def produces(self):
        return self._produces

    def get_path_parameter_types(self):
        types = {}
        path_parameters = (p for p in self.parameters if p["in"] == "path")
        for path_defn in path_parameters:
            if path_defn.get("type") == "string" and path_defn.get("format") == "path":
                # path is special case for type 'string'
                path_type = "path"
            else:
                path_type = path_defn.get("type")
            types[path_defn["name"]] = path_type
        return types

    def with_definitions(self, schema):
        if "schema" in schema:
            schema["schema"]["definitions"] = self.definitions
        return schema

    def response_schema(self, status_code=None, content_type=None):
        response_definition = self.response_definition(status_code, content_type)
        return self.with_definitions(response_definition.get("schema", {}))

    def example_response(self, status_code=None, *args, **kwargs):
        """
        Returns example response from spec
        """
        # simply use the first/lowest status code, this is probably 200 or 201
        status_code = status_code or sorted(self._responses.keys())[0]
        examples_path = [str(status_code), "examples"]
        schema_example_path = [str(status_code), "schema", "example"]
        schema_path = [str(status_code), "schema"]

        try:
            status_code = int(status_code)
        except ValueError:
            status_code = 200
        try:
            return (
                list(deep_get(self._responses, examples_path).values())[0],
                status_code,
            )
        except KeyError:
            pass
        try:
            return (deep_get(self._responses, schema_example_path), status_code)
        except KeyError:
            pass

        try:
            return (
                self._nested_example(deep_get(self._responses, schema_path)),
                status_code,
            )
        except KeyError:
            return (None, status_code)

    def _nested_example(self, schema):
        try:
            return schema["example"]
        except KeyError:
            pass
        try:
            # Recurse if schema is an object
            return {
                key: self._nested_example(value)
                for (key, value) in schema["properties"].items()
            }
        except KeyError:
            pass
        try:
            # Recurse if schema is an array
            return [self._nested_example(schema["items"])]
        except KeyError:
            raise

    def body_name(self, content_type: str = None) -> str:
        return self.body_definition(content_type).get("name", "body")

    def body_schema(self, content_type: str = None) -> dict:
        """
        The body schema definition for this operation.
        """
        body_definition = self.body_definition(content_type)
        return self.with_definitions(body_definition).get("schema", {})

    def body_definition(self, content_type: str = None) -> dict:
        """
        The body complete definition for this operation.

        **There can be one "body" parameter at most.**
        """
        return self.request_body

    def _transform_json(self, body_parameter: dict) -> dict:
        """Translate Swagger2 json parameters into OpenAPI 3 jsonschema spec."""
        nullable = body_parameter.get("x-nullable")
        if nullable is not None:
            body_parameter["schema"]["nullable"] = nullable
        return body_parameter

    def _transform_form(self, form_parameters: t.List[dict]) -> dict:
        """Translate Swagger2 form parameters into OpenAPI 3 jsonschema spec."""
        properties = {}
        defaults = {}
        required = []
        encoding = {}

        for param in form_parameters:
            prop = {}

            if param["type"] == "file":
                prop.update(
                    {
                        "type": "string",
                        "format": "binary",
                    }
                )
            else:
                prop["type"] = param["type"]

                format_ = param.get("format")
                if format_ is not None:
                    prop["format"] = format_

            default = param.get("default")
            if default is not None:
                prop["default"] = default
                defaults[param["name"]] = default

            nullable = param.get("x-nullable")
            if nullable is not None:
                prop["nullable"] = nullable

            if param["type"] == "array":
                prop["items"] = param.get("items", {})

                collection_format = param.get("collectionFormat", "csv")
                try:
                    encoding[param["name"]] = COLLECTION_FORMAT_MAPPING[
                        collection_format
                    ]
                except KeyError:
                    raise InvalidSpecification(
                        f"The collection format ({collection_format}) is not supported by "
                        f"Connexion as it cannot be mapped to OpenAPI 3."
                    )

            properties[param["name"]] = prop

            if param.get("required", False):
                required.append(param["name"])

        definition: t.Dict[str, t.Any] = {
            "schema": {
                "type": "object",
                "properties": properties,
                "required": required,
            }
        }
        if defaults:
            definition["schema"]["default"] = defaults
        if encoding:
            definition["encoding"] = encoding

        return definition



import asyncio
import functools
import logging
import typing as t

from starlette.concurrency import run_in_threadpool
from starlette.exceptions import HTTPException
from starlette.middleware.exceptions import (
    ExceptionMiddleware as StarletteExceptionMiddleware,
)
from starlette.requests import Request as StarletteRequest
from starlette.responses import Response as StarletteResponse
from starlette.types import ASGIApp, Receive, Scope, Send

from connexion import http_facts
from connexion.exceptions import InternalServerError, ProblemException, problem
from connexion.lifecycle import ConnexionRequest, ConnexionResponse
from connexion.types import MaybeAwaitable

logger = logging.getLogger(__name__)


def connexion_wrapper(
    handler: t.Callable[
        [ConnexionRequest, Exception], MaybeAwaitable[ConnexionResponse]
    ]
) -> t.Callable[[StarletteRequest, Exception], t.Awaitable[StarletteResponse]]:
    """Wrapper that translates Starlette requests to Connexion requests before passing
    them to the error handler, and translates the returned Connexion responses to
    Starlette responses."""

    @functools.wraps(handler)
    async def wrapper(request: StarletteRequest, exc: Exception) -> StarletteResponse:
        request = ConnexionRequest.from_starlette_request(request)

        if asyncio.iscoroutinefunction(handler):
            response = await handler(request, exc)  # type: ignore
        else:
            response = await run_in_threadpool(handler, request, exc)

        while asyncio.iscoroutine(response):
            response = await response

        return StarletteResponse(
            content=response.body,
            status_code=response.status_code,
            media_type=response.mimetype,
            headers=response.headers,
        )

    return wrapper


class ExceptionMiddleware(StarletteExceptionMiddleware):
    """Subclass of starlette ExceptionMiddleware to change handling of HTTP exceptions to
    existing connexion behavior."""

    def __init__(self, next_app: ASGIApp):
        super().__init__(next_app)
        self.add_exception_handler(ProblemException, self.problem_handler)  # type: ignore
        self.add_exception_handler(Exception, self.common_error_handler)

    def add_exception_handler(
        self,
        exc_class_or_status_code: t.Union[int, t.Type[Exception]],
        handler: t.Callable[[ConnexionRequest, Exception], StarletteResponse],
    ) -> None:
        super().add_exception_handler(
            exc_class_or_status_code, handler=connexion_wrapper(handler)
        )

    @staticmethod
    def problem_handler(_request: ConnexionRequest, exc: ProblemException):
        """Default handler for Connexion ProblemExceptions"""
        logger.error("%r", exc)
        return exc.to_problem()

    @staticmethod
    @connexion_wrapper
    def http_exception(
        _request: StarletteRequest, exc: HTTPException, **kwargs
    ) -> StarletteResponse:
        """Default handler for Starlette HTTPException"""
        logger.error("%r", exc)
        return problem(
            title=http_facts.HTTP_STATUS_CODES.get(exc.status_code),
            detail=exc.detail,
            status=exc.status_code,
            headers=exc.headers,
        )

    @staticmethod
    def common_error_handler(
        _request: StarletteRequest, exc: Exception
    ) -> ConnexionResponse:
        """Default handler for any unhandled Exception"""
        logger.error("%r", exc, exc_info=exc)
        return InternalServerError().to_problem()

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        await super().__call__(scope, receive, send)



"""
This module defines a SecurityHandlerFactory which supports the creation of
SecurityHandler instances for different security schemes.

It also exposes a `SECURITY_HANDLERS` dictionary which maps security scheme
types to SecurityHandler classes. This dictionary can be used to register
custom SecurityHandler classes for custom security schemes, or to overwrite
existing SecurityHandler classes.
This can be done by supplying a value for `security_map` argument of the
SecurityHandlerFactory.

Swagger 2.0 lets you define the following authentication types for an API:

- Basic authentication
- API key (as a header or a query string parameter)
- OAuth 2 common flows (authorization code, implicit, resource owner password credentials, client credentials)


Changes from OpenAPI 2.0 to OpenAPI 3.0
If you used OpenAPI 2.0 before, here is a summary of changes to help you get started with OpenAPI 3.0:
- securityDefinitions were renamed to securitySchemes and moved inside components.
- type: basic was replaced with type: http and scheme: basic.
- The new type: http is an umbrella type for all HTTP security schemes, including Basic, Bearer and other,
and the scheme keyword indicates the scheme type.
- API keys can now be sent in: cookie.
- Added support for OpenID Connect Discovery (type: openIdConnect).
- OAuth 2 security schemes can now define multiple flows.
- OAuth 2 flows were renamed to match the OAuth 2 Specification: accessCode is now authorizationCode,
and application is now clientCredentials.


OpenAPI uses the term security scheme for authentication and authorization schemes.
OpenAPI 3.0 lets you describe APIs protected using the following security schemes:

- HTTP authentication schemes (they use the Authorization header):
    - Basic
    - Bearer
    - other HTTP schemes as defined by RFC 7235 and HTTP Authentication Scheme Registry
- API keys in headers, query string or cookies
    - Cookie authentication
- OAuth 2
- OpenID Connect Discovery

"""

import asyncio
import base64
import http.cookies
import logging
import os
import typing as t

import httpx

from connexion.decorators.parameter import inspect_function_arguments
from connexion.exceptions import OAuthProblem, OAuthResponseProblem, OAuthScopeProblem
from connexion.lifecycle import ConnexionRequest
from connexion.utils import get_function_from_name

logger = logging.getLogger(__name__)


NO_VALUE = object()
"""Sentinel value to indicate that no security credentials were found."""


class AbstractSecurityHandler:

    required_scopes_kw = "required_scopes"
    client = None
    security_definition_key: str
    """The key which contains the value for the function name to resolve."""
    environ_key: str
    """The name of the environment variable that can be used alternatively for the function name."""

    def get_fn(self, security_scheme, required_scopes):
        """Returns the handler function"""
        security_func = self._resolve_func(security_scheme)
        if not security_func:
            logger.warning("... %s missing", self.security_definition_key)
            return None

        return self._get_verify_func(security_func)

    @classmethod
    def _get_function(
        cls,
        security_definition: dict,
        security_definition_key: str,
        environ_key: str,
        default: t.Optional[t.Callable] = None,
    ):
        """
        Return function by getting its name from security_definition or environment variable

        :param security_definition: Security Definition (scheme) from the spec.
        :param security_definition_key: The key which contains the value for the function name to resolve.
        :param environ_key: The name of the environment variable that can be used alternatively for the function name.
        :param default: The default to use in case the function cannot be found based on the security_definition_key or the environ_key
        """
        func_name = security_definition.get(security_definition_key) or os.environ.get(
            environ_key
        )
        if func_name:
            return get_function_from_name(func_name)
        return default

    def _generic_check(self, func, exception_msg):
        need_to_add_required_scopes = self._need_to_add_scopes(func)

        async def wrapper(request, *args, required_scopes=None):
            kwargs = {}
            if need_to_add_required_scopes:
                kwargs[self.required_scopes_kw] = required_scopes
            token_info = func(*args, **kwargs)
            while asyncio.iscoroutine(token_info):
                token_info = await token_info
            if token_info is NO_VALUE:
                return NO_VALUE
            if token_info is None:
                raise OAuthResponseProblem(detail=exception_msg)
            return token_info

        return wrapper

    @staticmethod
    def get_auth_header_value(request):
        """
        Return Authorization type and value if any.
        If not Authorization, return (None, None)
        Raise OAuthProblem for invalid Authorization header
        """
        authorization = request.headers.get("Authorization")
        if not authorization:
            return None, None

        try:
            auth_type, value = authorization.split(maxsplit=1)
        except ValueError:
            raise OAuthProblem(detail="Invalid authorization header")
        return auth_type.lower(), value

    def _need_to_add_scopes(self, func):
        arguments, has_kwargs = inspect_function_arguments(func)
        need_required_scopes = has_kwargs or self.required_scopes_kw in arguments
        return need_required_scopes

    def _resolve_func(self, security_scheme):
        """
        Get the user function object based on the security scheme or the environment variable.

        :param security_scheme: Security Definition (scheme) from the spec.
        """
        return self._get_function(
            security_scheme, self.security_definition_key, self.environ_key
        )

    def _get_verify_func(self, function):
        """
        Wraps the user security function in a function that checks the request for the correct
        security credentials and calls the user function with the correct arguments.
        """
        return self._generic_check(function, "Provided authorization is not valid")


class BasicSecurityHandler(AbstractSecurityHandler):
    """
    Security Handler for
    - `type: basic` (Swagger 2), and
    - `type: http` and `scheme: basic` (OpenAPI 3)
    """

    security_definition_key = "x-basicInfoFunc"
    environ_key = "BASICINFO_FUNC"

    def _get_verify_func(self, basic_info_func):
        check_basic_info_func = self.check_basic_auth(basic_info_func)

        def wrapper(request):
            auth_type, user_pass = self.get_auth_header_value(request)
            if auth_type != "basic":
                return NO_VALUE

            try:
                username, password = (
                    base64.b64decode(user_pass).decode("latin1").split(":", 1)
                )
            except Exception:
                raise OAuthProblem(detail="Invalid authorization header")

            return check_basic_info_func(request, username, password)

        return wrapper

    def check_basic_auth(self, basic_info_func):
        return self._generic_check(
            basic_info_func, "Provided authorization is not valid"
        )


class BearerSecurityHandler(AbstractSecurityHandler):
    """
    Security Handler for HTTP Bearer authentication.
    """

    security_definition_key = "x-bearerInfoFunc"
    environ_key = "BEARERINFO_FUNC"

    def check_bearer_token(self, token_info_func):
        return self._generic_check(token_info_func, "Provided token is not valid")

    def _get_verify_func(self, token_info_func):
        """
        :param token_info_func: types.FunctionType
        :rtype: types.FunctionType
        """
        check_bearer_func = self.check_bearer_token(token_info_func)

        def wrapper(request):
            auth_type, token = self.get_auth_header_value(request)
            if auth_type != "bearer":
                return NO_VALUE
            return check_bearer_func(request, token)

        return wrapper


class ApiKeySecurityHandler(AbstractSecurityHandler):
    """
    Security Handler for API Keys.
    """

    security_definition_key = "x-apikeyInfoFunc"
    environ_key = "APIKEYINFO_FUNC"

    def get_fn(self, security_scheme, required_scopes):
        apikey_info_func = self._resolve_func(security_scheme)
        if not apikey_info_func:
            logger.warning("... %s missing", self.security_definition_key)
            return None

        return self._get_verify_func(
            apikey_info_func,
            security_scheme["in"],
            security_scheme["name"],
        )

    def _get_verify_func(self, api_key_info_func, loc, name):
        check_api_key_func = self.check_api_key(api_key_info_func)

        def wrapper(request: ConnexionRequest):
            if loc == "query":
                api_key = request.query_params.get(name)
            elif loc == "header":
                api_key = request.headers.get(name)
            elif loc == "cookie":
                cookie_list = request.headers.get("Cookie")
                api_key = self.get_cookie_value(cookie_list, name)
            else:
                return NO_VALUE

            if api_key is None:
                return NO_VALUE

            return check_api_key_func(request, api_key)

        return wrapper

    def check_api_key(self, api_key_info_func):
        return self._generic_check(api_key_info_func, "Provided apikey is not valid")

    @staticmethod
    def get_cookie_value(cookies, name):
        """
        Returns cookie value by its name. `None` if no such value.

        :param cookies: str: cookies raw data
        :param name: str: cookies key
        """
        cookie_parser = http.cookies.SimpleCookie()
        cookie_parser.load(str(cookies))
        try:
            return cookie_parser[name].value
        except KeyError:
            return None


class OAuthSecurityHandler(AbstractSecurityHandler):
    """
    Security Handler for the OAuth security scheme.
    """

    def get_fn(self, security_scheme, required_scopes):
        token_info_func = self.get_tokeninfo_func(security_scheme)
        scope_validate_func = self.get_scope_validate_func(security_scheme)
        if not token_info_func:
            logger.warning("... x-tokenInfoFunc missing")
            return None

        return self._get_verify_func(
            token_info_func, scope_validate_func, required_scopes
        )

    def get_tokeninfo_func(self, security_definition: dict) -> t.Optional[t.Callable]:
        """
        Gets the function for retrieving the token info.
        It is possible to specify a function or a URL. The function variant is
        preferred. If it is not found, the URL variant is used with the
        `get_token_info_remote` function.

        >>> get_tokeninfo_func({'x-tokenInfoFunc': 'foo.bar'})
        '<function foo.bar>'
        """
        token_info_func = self._get_function(
            security_definition, "x-tokenInfoFunc", "TOKENINFO_FUNC"
        )
        if token_info_func:
            return token_info_func

        token_info_url = security_definition.get("x-tokenInfoUrl") or os.environ.get(
            "TOKENINFO_URL"
        )
        if token_info_url:
            return self.get_token_info_remote(token_info_url)

        return None

    @classmethod
    def get_scope_validate_func(cls, security_definition):
        """
        Gets the function for validating the token scopes.
        If it is not found, the default `validate_scope` function is used.

        >>> get_scope_validate_func({'x-scopeValidateFunc': 'foo.bar'})
        '<function foo.bar>'
        """
        return cls._get_function(
            security_definition,
            "x-scopeValidateFunc",
            "SCOPEVALIDATE_FUNC",
            cls.validate_scope,
        )

    @staticmethod
    def validate_scope(required_scopes, token_scopes):
        """
        :param required_scopes: Scopes required to access operation
        :param token_scopes: Scopes granted by authorization server
        :rtype: bool
        """
        required_scopes = set(required_scopes)
        if isinstance(token_scopes, list):
            token_scopes = set(token_scopes)
        else:
            token_scopes = set(token_scopes.split())
        logger.debug("... Scopes required: %s", required_scopes)
        logger.debug("... Token scopes: %s", token_scopes)
        if not required_scopes <= token_scopes:
            logger.info(
                "... Token scopes (%s) do not match the scopes necessary to call endpoint (%s)."
                " Aborting with 403.",
                token_scopes,
                required_scopes,
            )
            return False
        return True

    def get_token_info_remote(self, token_info_url: str) -> t.Callable:
        """
        Return a function which will call `token_info_url` to retrieve token info.

        Returned function must accept oauth token in parameter.
        It must return a token_info dict in case of success, None otherwise.

        :param token_info_url: URL to get information about the token
        """

        async def wrapper(token):
            if self.client is None:
                self.client = httpx.AsyncClient()
            headers = {"Authorization": f"Bearer {token}"}
            token_request = await self.client.get(
                token_info_url, headers=headers, timeout=5
            )
            if token_request.status_code != 200:
                return
            return token_request.json()

        return wrapper

    def _get_verify_func(self, token_info_func, scope_validate_func, required_scopes):
        check_oauth_func = self.check_oauth_func(token_info_func, scope_validate_func)

        def wrapper(request):
            auth_type, token = self.get_auth_header_value(request)
            if auth_type != "bearer":
                return NO_VALUE

            return check_oauth_func(request, token, required_scopes=required_scopes)

        return wrapper

    def check_oauth_func(self, token_info_func, scope_validate_func):
        get_token_info = self._generic_check(
            token_info_func, "Provided token is not valid"
        )

        async def wrapper(request, token, required_scopes):
            token_info = await get_token_info(
                request, token, required_scopes=required_scopes
            )

            # Fallback to 'scopes' for backward compatibility
            token_scopes = token_info.get("scope", token_info.get("scopes", ""))

            validation = scope_validate_func(required_scopes, token_scopes)
            while asyncio.iscoroutine(validation):
                validation = await validation
            if not validation:
                raise OAuthScopeProblem(
                    required_scopes=required_scopes,
                    token_scopes=token_scopes,
                )

            return token_info

        return wrapper


SECURITY_HANDLERS = {
    # Swagger 2: `type: basic`
    # OpenAPI 3: `type: http` and `scheme: basic`
    "basic": BasicSecurityHandler,
    # Swagger 2 and OpenAPI 3
    "apiKey": ApiKeySecurityHandler,
    "oauth2": OAuthSecurityHandler,
    # OpenAPI 3: http schemes
    "bearer": BearerSecurityHandler,
}


class SecurityHandlerFactory:
    """
    A factory class for parsing security schemes and returning the appropriate
    security handler.

    By default, it will use the built-in security handlers specified in the
    SECURITY_HANDLERS dict, but you can also pass in your own security handlers
    to override the built-in ones.
    """

    def __init__(
        self,
        security_handlers: t.Optional[dict] = None,
    ) -> None:
        self.security_handlers = SECURITY_HANDLERS.copy()
        if security_handlers is not None:
            self.security_handlers.update(security_handlers)

    def parse_security_scheme(
        self,
        security_scheme: dict,
        required_scopes: t.List[str],
    ) -> t.Optional[t.Callable]:
        """Parses the security scheme and returns the function for verifying it.

        :param security_scheme: The security scheme from the spec.
        :param required_scopes: List of scopes for this security scheme.
        """
        security_type = security_scheme["type"]
        if security_type in ("basic", "oauth2"):
            security_handler = self.security_handlers[security_type]
            return security_handler().get_fn(security_scheme, required_scopes)

        # OpenAPI 3.0.0
        elif security_type == "http":
            scheme = security_scheme["scheme"].lower()
            if scheme in self.security_handlers:
                security_handler = self.security_handlers[scheme]
                return security_handler().get_fn(security_scheme, required_scopes)
            else:
                logger.warning("... Unsupported http authorization scheme %s", scheme)
                return None

        elif security_type == "apiKey":
            scheme = security_scheme.get("x-authentication-scheme", "").lower()
            if scheme == "bearer":
                return BearerSecurityHandler().get_fn(security_scheme, required_scopes)
            else:
                security_handler = self.security_handlers["apiKey"]
                return security_handler().get_fn(security_scheme, required_scopes)

        # Custom security handler
        elif (scheme := security_scheme["scheme"].lower()) in self.security_handlers:
            security_handler = self.security_handlers[scheme]
            return security_handler().get_fn(security_scheme, required_scopes)

        else:
            logger.warning(
                "... Unsupported security scheme type %s",
                security_type,
            )
            return None

    @staticmethod
    async def security_passthrough(request):
        """Used when no security is required for the operation.

        Equivalent OpenAPI snippet:

        .. code-block:: yaml

            /helloworld
              get:
                security: []   # No security
                ...
        """
        return request

    @staticmethod
    def verify_none(request):
        """Used for optional security.

        Equivalent OpenAPI snippet:

        .. code-block:: yaml

            security:
              - {}  # <--
              - myapikey: []
        """
        return {}

    def verify_multiple_schemes(self, schemes):
        """
        Verifies multiple authentication schemes in AND fashion.
        If any scheme fails, the entire authentication fails.

        :param schemes: mapping scheme_name to auth function
        :type schemes: dict
        :rtype: types.FunctionType
        """

        async def wrapper(request):
            token_info = {}
            for scheme_name, func in schemes.items():
                result = func(request)
                while asyncio.iscoroutine(result):
                    result = await result
                if result is NO_VALUE:
                    return NO_VALUE
                token_info[scheme_name] = result

            return token_info

        return wrapper

    @classmethod
    def verify_security(cls, auth_funcs):
        async def verify_fn(request):
            token_info = NO_VALUE
            errors = []
            for func in auth_funcs:
                try:
                    token_info = func(request)
                    while asyncio.iscoroutine(token_info):
                        token_info = await token_info
                    if token_info is not NO_VALUE:
                        break
                except Exception as err:
                    errors.append(err)

            else:
                if errors != []:
                    cls._raise_most_specific(errors)
                else:
                    logger.info("... No auth provided. Aborting with 401.")
                    raise OAuthProblem(detail="No authorization token provided")

            request.context.update(
                {
                    # Fallback to 'uid' for backward compatibility
                    "user": token_info.get("sub", token_info.get("uid")),
                    "token_info": token_info,
                }
            )

        return verify_fn

    @staticmethod
    def _raise_most_specific(exceptions: t.List[Exception]) -> None:
        """Raises the most specific error from a list of exceptions by status code.

        The status codes are expected to be either in the `code`
        or in the `status` attribute of the exceptions.

        The order is as follows:
            - 403: valid credentials but not enough privileges
            - 401: no or invalid credentials
            - for other status codes, the smallest one is selected

        :param errors: List of exceptions.
        :type errors: t.List[Exception]
        """
        if not exceptions:
            return
        # We only use status code attributes from exceptions
        # We use 600 as default because 599 is highest valid status code
        status_to_exc = {
            getattr(exc, "status_code", getattr(exc, "status", 600)): exc
            for exc in exceptions
        }
        if 403 in status_to_exc:
            raise status_to_exc[403]
        elif 401 in status_to_exc:
            raise status_to_exc[401]
        else:
            lowest_status_code = min(status_to_exc)
            raise status_to_exc[lowest_status_code]



import typing as t
from contextvars import ContextVar

import starlette.convertors
from starlette.routing import Router
from starlette.types import ASGIApp, Receive, Scope, Send

from connexion.frameworks import starlette as starlette_utils
from connexion.middleware.abstract import (
    ROUTING_CONTEXT,
    AbstractRoutingAPI,
    SpecMiddleware,
)
from connexion.operations import AbstractOperation
from connexion.resolver import Resolver
from connexion.spec import Specification

_scope: ContextVar[dict] = ContextVar("SCOPE")


class RoutingOperation:
    def __init__(self, operation_id: t.Optional[str], next_app: ASGIApp) -> None:
        self.operation_id = operation_id
        self.next_app = next_app

    @classmethod
    def from_operation(cls, operation: AbstractOperation, next_app: ASGIApp):
        return cls(operation.operation_id, next_app)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """Attach operation to scope and pass it to the next app"""
        original_scope = _scope.get()
        # Pass resolved path params along
        original_scope.setdefault("path_params", {}).update(
            scope.get("path_params", {})
        )

        api_base_path = scope.get("root_path", "")[
            len(original_scope.get("root_path", "")) :
        ]

        extensions = original_scope.setdefault("extensions", {})
        connexion_routing = extensions.setdefault(ROUTING_CONTEXT, {})
        connexion_routing.update(
            {"api_base_path": api_base_path, "operation_id": self.operation_id}
        )
        await self.next_app(original_scope, receive, send)


class RoutingAPI(AbstractRoutingAPI):
    def __init__(
        self,
        specification: Specification,
        *,
        next_app: ASGIApp,
        base_path: t.Optional[str] = None,
        arguments: t.Optional[dict] = None,
        resolver: t.Optional[Resolver] = None,
        resolver_error_handler: t.Optional[t.Callable] = None,
        debug: bool = False,
        **kwargs,
    ) -> None:
        """API implementation on top of Starlette Router for Connexion middleware."""
        self.next_app = next_app
        self.router = Router(default=RoutingOperation(None, next_app))

        super().__init__(
            specification,
            base_path=base_path,
            arguments=arguments,
            resolver=resolver,
            resolver_error_handler=resolver_error_handler,
            debug=debug,
            **kwargs,
        )

    def make_operation(self, operation: AbstractOperation) -> RoutingOperation:
        return RoutingOperation.from_operation(operation, next_app=self.next_app)

    @staticmethod
    def _framework_path_and_name(
        operation: AbstractOperation, path: str
    ) -> t.Tuple[str, str]:
        types = operation.get_path_parameter_types()
        starlette_path = starlette_utils.starlettify_path(path, types)
        return starlette_path, starlette_path

    def _add_operation_internal(
        self, method: str, path: str, operation: RoutingOperation, name: str = None
    ) -> None:
        self.router.add_route(path, operation, methods=[method])


class RoutingMiddleware(SpecMiddleware):
    def __init__(self, app: ASGIApp) -> None:
        """Middleware that resolves the Operation for an incoming request and attaches it to the
        scope.

        :param app: app to wrap in middleware.
        """
        self.app = app
        # Pass unknown routes to next app
        self.router = Router(default=RoutingOperation(None, self.app))
        starlette.convertors.register_url_convertor(
            "float", starlette_utils.FloatConverter()
        )
        starlette.convertors.register_url_convertor(
            "int", starlette_utils.IntegerConverter()
        )

    def add_api(
        self,
        specification: Specification,
        base_path: t.Optional[str] = None,
        arguments: t.Optional[dict] = None,
        **kwargs,
    ) -> None:
        """Add an API to the router based on a OpenAPI spec.

        :param specification: OpenAPI spec.
        :param base_path: Base path where to add this API.
        :param arguments: Jinja arguments to replace in the spec.
        """
        api = RoutingAPI(
            specification,
            base_path=base_path,
            arguments=arguments,
            next_app=self.app,
            **kwargs,
        )

        # If an API with the same base_path was already registered, chain the new API as its
        # default. This way, if no matching route is found on the first API, the request is
        # forwarded to the new API.
        for route in self.router.routes:
            if (
                isinstance(route, starlette.routing.Mount)
                and route.path == api.base_path
            ):
                route.app.default = api.router

        self.router.mount(api.base_path, app=api.router)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """Route request to matching operation, and attach it to the scope before calling the
        next app."""
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        _scope.set(scope.copy())  # type: ignore

        await self.router(scope, receive, send)



from connexion.datastructures import MediaTypeDict

from .abstract import (  # NOQA
    AbstractRequestBodyValidator,
    AbstractResponseBodyValidator,
)
from .form_data import FormDataValidator, MultiPartFormDataValidator
from .json import DefaultsJSONRequestBodyValidator  # NOQA
from .json import (
    JSONRequestBodyValidator,
    JSONResponseBodyValidator,
    TextResponseBodyValidator,
)
from .parameter import ParameterValidator

VALIDATOR_MAP = {
    "parameter": ParameterValidator,
    "body": MediaTypeDict(
        {
            "*/*json": JSONRequestBodyValidator,
            "application/x-www-form-urlencoded": FormDataValidator,
            "multipart/form-data": MultiPartFormDataValidator,
        }
    ),
    "response": MediaTypeDict(
        {
            "*/*json": JSONResponseBodyValidator,
            "text/plain": TextResponseBodyValidator,
        }
    ),
}



"""
Validation Middleware.
"""
import logging
import typing as t

from starlette.types import ASGIApp, Receive, Scope, Send

from connexion import utils
from connexion.datastructures import MediaTypeDict
from connexion.exceptions import UnsupportedMediaTypeProblem
from connexion.middleware.abstract import RoutedAPI, RoutedMiddleware
from connexion.operations import AbstractOperation
from connexion.validators import VALIDATOR_MAP

logger = logging.getLogger("connexion.middleware.validation")


class RequestValidationOperation:
    def __init__(
        self,
        next_app: ASGIApp,
        *,
        operation: AbstractOperation,
        strict_validation: bool = False,
        validator_map: t.Optional[dict] = None,
    ) -> None:
        self.next_app = next_app
        self._operation = operation
        self.strict_validation = strict_validation
        self._validator_map = VALIDATOR_MAP.copy()
        self._validator_map.update(validator_map or {})

    def extract_content_type(
        self, headers: t.List[t.Tuple[bytes, bytes]]
    ) -> t.Tuple[str, str]:
        """Extract the mime type and encoding from the content type headers.

        :param headers: Headers from ASGI scope

        :return: A tuple of mime type, encoding
        """
        content_type = utils.extract_content_type(headers)
        mime_type, encoding = utils.split_content_type(content_type)
        if mime_type is None:
            # Content-type header is not required. Take a best guess.
            try:
                mime_type = self._operation.consumes[0]
            except IndexError:
                mime_type = "application/octet-stream"
        if encoding is None:
            encoding = "utf-8"

        return mime_type, encoding

    def validate_mime_type(self, mime_type: str) -> None:
        """Validate the mime type against the spec if it defines which mime types are accepted.

        :param mime_type: mime type from content type header
        """
        if not self._operation.consumes:
            return

        # Convert to MediaTypeDict to handle media-ranges
        media_type_dict = MediaTypeDict(
            [(c.lower(), None) for c in self._operation.consumes]
        )
        if mime_type.lower() not in media_type_dict:
            raise UnsupportedMediaTypeProblem(
                detail=f"Invalid Content-type ({mime_type}), "
                f"expected {self._operation.consumes}"
            )

    @property
    def security_query_params(self) -> t.List[str]:
        """Get the names of query parameters that are used for security."""
        if not hasattr(self, "_security_query_params"):
            security_query_params: t.List[str] = []
            if self._operation.security is None:
                self._security_query_params = security_query_params
                return self._security_query_params

            for security_req in self._operation.security:
                for scheme_name in security_req:
                    security_scheme = self._operation.security_schemes[scheme_name]

                    if (
                        security_scheme["type"] == "apiKey"
                        and security_scheme["in"] == "query"
                    ):
                        # Only query parameters need to be considered for strict_validation
                        security_query_params.append(security_scheme["name"])
            self._security_query_params = security_query_params

        return self._security_query_params

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        # Validate parameters & headers
        uri_parser_class = self._operation._uri_parser_class
        uri_parser = uri_parser_class(
            self._operation.parameters, self._operation.body_definition()
        )
        parameter_validator_cls = self._validator_map["parameter"]
        parameter_validator = parameter_validator_cls(  # type: ignore
            self._operation.parameters,
            uri_parser=uri_parser,
            strict_validation=self.strict_validation,
            security_query_params=self.security_query_params,
        )
        parameter_validator.validate(scope)

        # Extract content type
        headers = scope["headers"]
        mime_type, encoding = self.extract_content_type(headers)
        self.validate_mime_type(mime_type)

        # Validate body
        schema = self._operation.body_schema(mime_type)
        if schema:
            try:
                body_validator = self._validator_map["body"][mime_type]  # type: ignore
            except KeyError:
                logging.info(
                    f"Skipping validation. No validator registered for content type: "
                    f"{mime_type}."
                )
            else:
                validator = body_validator(
                    schema=schema,
                    required=self._operation.request_body.get("required", False),
                    nullable=utils.is_nullable(
                        self._operation.body_definition(mime_type)
                    ),
                    encoding=encoding,
                    strict_validation=self.strict_validation,
                    uri_parser=self._operation.uri_parser_class(
                        self._operation.parameters, self._operation.body_definition()
                    ),
                )
                receive = await validator.wrap_receive(receive, scope=scope)

        await self.next_app(scope, receive, send)


class RequestValidationAPI(RoutedAPI[RequestValidationOperation]):
    """Validation API."""

    def __init__(
        self,
        *args,
        strict_validation=False,
        validator_map=None,
        uri_parser_class=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.validator_map = validator_map

        logger.debug("Strict Request Validation: %s", str(strict_validation))
        self.strict_validation = strict_validation

        self.uri_parser_class = uri_parser_class

        self.add_paths()

    def make_operation(
        self, operation: AbstractOperation
    ) -> RequestValidationOperation:
        return RequestValidationOperation(
            self.next_app,
            operation=operation,
            strict_validation=self.strict_validation,
            validator_map=self.validator_map,
        )


class RequestValidationMiddleware(RoutedMiddleware[RequestValidationAPI]):
    """Middleware for validating requests according to the API contract."""

    api_cls = RequestValidationAPI


class MissingValidationOperation(Exception):
    """Missing validation operation"""



import json
import logging
import re
import typing as t
from contextvars import ContextVar

from starlette.requests import Request as StarletteRequest
from starlette.responses import RedirectResponse
from starlette.responses import Response as StarletteResponse
from starlette.routing import Router
from starlette.staticfiles import StaticFiles
from starlette.templating import Jinja2Templates
from starlette.types import ASGIApp, Receive, Scope, Send

from connexion.jsonifier import Jsonifier
from connexion.middleware import SpecMiddleware
from connexion.middleware.abstract import AbstractSpecAPI
from connexion.options import SwaggerUIConfig, SwaggerUIOptions
from connexion.spec import Specification
from connexion.utils import yamldumper

logger = logging.getLogger("connexion.middleware.swagger_ui")


_original_scope: ContextVar[Scope] = ContextVar("SCOPE")


class SwaggerUIAPI(AbstractSpecAPI):
    def __init__(
        self,
        *args,
        default: ASGIApp,
        swagger_ui_options: t.Optional[SwaggerUIOptions] = None,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        self.router = Router(default=default)
        self.options = SwaggerUIConfig(
            swagger_ui_options, oas_version=self.specification.version
        )

        if self.options.openapi_spec_available:
            self.add_openapi_json()
            self.add_openapi_yaml()

        if self.options.swagger_ui_available:
            self.add_swagger_ui()

        self._templates = Jinja2Templates(
            directory=str(self.options.swagger_ui_template_dir)
        )

    @staticmethod
    def normalize_string(string):
        return re.sub(r"[^a-zA-Z0-9]", "_", string.strip("/"))

    def _base_path_for_prefix(self, request: StarletteRequest) -> str:
        """
        returns a modified basePath which includes the incoming root_path.
        """
        return request.scope.get("root_path", "").rstrip("/")

    def _spec_for_prefix(self, request):
        """
        returns a spec with a modified basePath / servers block
        which corresponds to the incoming request path.
        This is needed when behind a path-altering reverse proxy.
        """
        base_path = self._base_path_for_prefix(request)
        return self.specification.with_base_path(base_path).raw

    def add_openapi_json(self):
        """
        Adds openapi json to {base_path}/openapi.json
             (or {base_path}/swagger.json for swagger2)
        """
        logger.info(
            "Adding spec json: %s%s", self.base_path, self.options.openapi_spec_path
        )
        self.router.add_route(
            methods=["GET"],
            path=self.options.openapi_spec_path,
            endpoint=self._get_openapi_json,
        )

    def add_openapi_yaml(self):
        """
        Adds openapi json to {base_path}/openapi.json
             (or {base_path}/swagger.json for swagger2)
        """
        if not self.options.openapi_spec_path.endswith("json"):
            return

        openapi_spec_path_yaml = self.options.openapi_spec_path[: -len("json")] + "yaml"
        logger.debug("Adding spec yaml: %s/%s", self.base_path, openapi_spec_path_yaml)
        self.router.add_route(
            methods=["GET"],
            path=openapi_spec_path_yaml,
            endpoint=self._get_openapi_yaml,
        )

    async def _get_openapi_json(self, request):
        # Yaml parses datetime objects when loading the spec, so we need our custom jsonifier to dump it
        jsonifier = Jsonifier()

        return StarletteResponse(
            content=jsonifier.dumps(self._spec_for_prefix(request)),
            status_code=200,
            media_type="application/json",
        )

    async def _get_openapi_yaml(self, request):
        return StarletteResponse(
            content=yamldumper(self._spec_for_prefix(request)),
            status_code=200,
            media_type="text/yaml",
        )

    def add_swagger_ui(self):
        """
        Adds swagger ui to {base_path}/ui/
        """
        console_ui_path = self.options.swagger_ui_path.strip().rstrip("/")
        logger.debug("Adding swagger-ui: %s%s/", self.base_path, console_ui_path)

        for path in (
            console_ui_path + "/",
            console_ui_path + "/index.html",
        ):
            self.router.add_route(
                methods=["GET"], path=path, endpoint=self._get_swagger_ui_home
            )

        if self.options.swagger_ui_config:
            self.router.add_route(
                methods=["GET"],
                path=console_ui_path + "/swagger-ui-config.json",
                endpoint=self._get_swagger_ui_config,
            )

        # we have to add an explicit redirect instead of relying on the
        # normalize_path_middleware because we also serve static files
        # from this dir (below)

        async def redirect(request):
            url = request.scope.get("root_path", "").rstrip("/")
            url += console_ui_path
            url += "/"
            return RedirectResponse(url=url)

        self.router.add_route(methods=["GET"], path=console_ui_path, endpoint=redirect)

        # this route will match and get a permission error when trying to
        # serve index.html, so we add the redirect above.
        self.router.mount(
            path=console_ui_path,
            app=StaticFiles(directory=str(self.options.swagger_ui_template_dir)),
            name="swagger_ui_static",
        )

    async def _get_swagger_ui_home(self, req):
        base_path = self._base_path_for_prefix(req)
        template_variables = {
            "request": req,
            "openapi_spec_url": (base_path + self.options.openapi_spec_path),
            **self.options.swagger_ui_template_arguments,
        }
        if self.options.swagger_ui_config:
            template_variables["configUrl"] = "swagger-ui-config.json"

        return self._templates.TemplateResponse("index.j2", template_variables)

    async def _get_swagger_ui_config(self, request):
        return StarletteResponse(
            status_code=200,
            media_type="application/json",
            content=json.dumps(self.options.swagger_ui_config),
        )


class SwaggerUIMiddleware(SpecMiddleware):
    def __init__(self, app: ASGIApp) -> None:
        """Middleware that hosts a swagger UI.

        :param app: app to wrap in middleware.
        """
        self.app = app
        # Set default to pass unknown routes to next app
        self.router = Router(default=self.default_fn)

    def add_api(
        self,
        specification: Specification,
        base_path: t.Optional[str] = None,
        arguments: t.Optional[dict] = None,
        **kwargs
    ) -> None:
        """Add an API to the router based on a OpenAPI spec.

        :param specification: OpenAPI spec.
        :param base_path: Base path where to add this API.
        :param arguments: Jinja arguments to replace in the spec.
        """
        api = SwaggerUIAPI(
            specification,
            base_path=base_path,
            arguments=arguments,
            default=self.default_fn,
            **kwargs
        )
        self.router.mount(api.base_path, app=api.router)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        _original_scope.set(scope.copy())  # type: ignore
        await self.router(scope, receive, send)

    async def default_fn(self, _scope: Scope, receive: Receive, send: Send) -> None:
        """
        Callback to call next app as default when no matching route is found.

        Unfortunately we cannot just pass the next app as default, since the router manipulates
        the scope when descending into mounts, losing information about the base path. Therefore,
        we use the original scope instead.

        This is caused by https://github.com/encode/starlette/issues/1336.
        """
        original_scope = _original_scope.get()
        await self.app(original_scope, receive, send)
